"""_summary_."""
from typing import List, Dict
from os.path import expanduser
import os
from lifespline_utils import utils
from lifespline_utils.error import LSUDuplicateConfigSectionsError
from requests_toolbelt import user_agent


def test_parse_files_invalid_files():
    """Verify if parsing at least one invalid file, the exception 
    LSUInvalidFileFormatException is thrown.
    """
    # TODO: add files with edge cases
    invalid_file_paths: List[str] = [
        "test/test_utils/invalid_setup.cfg",
    ]
    assert True


def test_parse_files_valid_files():
    """Verify if parsing only valid files, the exception 
    LSUInvalidFileFormatException is not thrown.
    """
    # TODO: add files with edge cases
    valid_file_paths: List[str] = [
        "test/test_utils/valid_setup.cfg",
    ]
    assert True


def test_parse_files_verify_correct_parse():
    """Verifies that valid cfg files are correctly parsed.
    """
    valid_file_paths: List[str] = [
        "test/test_utils/valid_setup.cfg",
    ]
    expected_output: Dict[str, Dict[str, str]] = {
        'section': {
            'key': 'value'
        }
    }
    assert True


def test_parse_files_verify_missing_section_header():
    """Verify that duplicate section headers will raise `lifespline_utils.error.LSUDuplicateConfigSectionsError` 
    """
    res: bool = False

    # read configuration files that have a duplicate section
    root: str = "test/test_config/"
    conflicting_setups: List[str] = [
        f"{root}setup1.cfg",
        f"{root}setup3.cfg",
    ]

    # verify that the correct error is raised
    try:
        utils.parse_files(*conflicting_setups)
    except LSUDuplicateConfigSectionsError as err:
        res = True

    assert res


def test_parse_files_verify_duplicate_sections():
    """Verify that duplicate section headers will raise `lifespline_utils.error.LSUDuplicateConfigSectionsError` 
    """
    res: bool = False

    # read configuration files that have a duplicate section
    root: str = "test/test_config/"
    conflicting_setups: List[str] = [
        f"{root}setup1.cfg",
        f"{root}setup3.cfg",
    ]

    # verify that the correct error is raised
    try:
        utils.parse_files(*conflicting_setups)
    except LSUDuplicateConfigSectionsError as err:
        res = True

    assert res


def test_parse_files_no_duplicate_sections():
    """Verify that duplicate section headers will raise `lifespline_utils.error.LSUDuplicateConfigSectionsError` 
    """
    res: bool = True

    # read configuration files that don't have a duplicate section
    root: str = "test/test_config/"
    conflicting_setups: List[str] = [
        f"{root}setup1.cfg",
        f"{root}setup2.cfg",
    ]

    # verify that no error is raised
    utils.parse_files(*conflicting_setups)

    assert res


def test_get_project_root_from_repo():
    """Verify if, after installing the package from the python package index, the project root is as expected.

    ```bash
    >>> from lifespline_utils.utils import get_path, get_project_root
    >>> get_project_root()
    '/home/dipm/src/lifespline-utils/'
    >>> get_path()
    '/home/dipm/src/lifespline-utils/.env/lib/python3.8/site-packages/lifespline_utils-0.0.8-py3.8.egg/lifespline_utils/utils.py'
    >>> get_path('README.md')
    '/home/dipm/src/lifespline-utils/README.md'
    >>> get_path('.bashrc')
    '/home/dipm/.bashrc'
    ```
    """
    assert True


def test_get_project_root_from_src():
    """Verify if, in debug mode, the project root is as expected.

    Running from local package source code
    ```bash
    >>> from lifespline_utils.utils import get_path, get_project_root
    >>> get_project_root()
    '/home/dipm/src/lifespline-utils/'
    >>> get_path()
    '/home/dipm/src/lifespline-utils/.env/lib/python3.8/site-packages/lifespline_utils-0.0.8-py3.8.egg/lifespline_utils/utils.py'
    >>> get_path('README.md')
    '/home/dipm/src/lifespline-utils/README.md'
    >>> get_path('.bashrc')
    '/home/dipm/.bashrc'
    ```
    """
    assert True


def test_get_project_root_from_build():
    """Verify if in pip-install mode, the project root is as expected.

    Running from local package build
    ```bash
    >>> from lifespline_utils.utils import get_path, get_project_root
    >>> get_project_root()
    '/home/dipm/src/lifespline-utils/'
    >>> get_path()
    '/home/dipm/src/lifespline-utils/.env/lib/python3.8/site-packages/lifespline_utils-0.0.8-py3.8.egg/lifespline_utils/utils.py'
    >>> get_path('README.md')
    '/home/dipm/src/lifespline-utils/README.md'
    >>> get_path('.bashrc')
    '/home/dipm/.bashrc'
    ```
    """
    assert True


def test_get_path_from_repo():
    """Verify if, after installing the package from the python package index, the project root is as expected.

    ```bash
    >>> from lifespline_utils.utils import get_path, get_project_root
    >>> get_project_root()
    '/home/dipm/src/lifespline-utils/'
    >>> get_path()
    '/home/dipm/src/lifespline-utils/.env/lib/python3.8/site-packages/lifespline_utils-0.0.8-py3.8.egg/lifespline_utils/utils.py'
    >>> get_path('README.md')
    '/home/dipm/src/lifespline-utils/README.md'
    >>> get_path('.bashrc')
    '/home/dipm/.bashrc'
    ```
    """
    assert True


def test_get_path_from_src():
    """Verify if, in debug mode, the project root is as expected.

    Running from local package source code
    ```bash
    >>> from lifespline_utils.utils import get_path, get_project_root
    >>> get_project_root()
    '/home/dipm/src/lifespline-utils/'
    >>> get_path()
    '/home/dipm/src/lifespline-utils/.env/lib/python3.8/site-packages/lifespline_utils-0.0.8-py3.8.egg/lifespline_utils/utils.py'
    >>> get_path('README.md')
    '/home/dipm/src/lifespline-utils/README.md'
    >>> get_path('.bashrc')
    '/home/dipm/.bashrc'
    ```
    """
    assert True


def test_get_path_from_build():
    """Verify if in pip-install mode, the project root is as expected.

    Running from local package build
    ```bash
    >>> from lifespline_utils.utils import get_path, get_project_root
    >>> get_project_root()
    '/home/dipm/src/lifespline-utils/'
    >>> get_path()
    '/home/dipm/src/lifespline-utils/.env/lib/python3.8/site-packages/lifespline_utils-0.0.8-py3.8.egg/lifespline_utils/utils.py'
    >>> get_path('README.md')
    '/home/dipm/src/lifespline-utils/README.md'
    >>> get_path('.bashrc')
    '/home/dipm/.bashrc'
    ```
    """
    assert True



"""Logging."""
import logging
from typing import Dict, List
import json

_log_format: str = "%(message)s"

logging.basicConfig(
    format=_log_format,
    level=logging.INFO,
)

logger = logging.getLogger("root")

class CLI:
    """CLI log styling
    """
    pass

class API:
    """API log styling
    """
    pass

class LogColour:
    """
    ANSI escape code: \033
    ANSI format: \033[{style};{foreground};{background}m
    https://ozzmaker.com/add-colour-to-text-in-python/
    """
    ansi_code = '\033['
    fg = {
        "black" : '30',
        "red" : '31',
        "green" : '32',
        "yellow" : '33',
        "blue" : '34',
        "purple" : '35',
        "cyan" : '36',
        "white" : '37',
        "none": '0'
    }

    bg = {
        "black" : '40',
        "red" : '41',
        "green" : '42',
        "yellow" : '43',
        "blue" : '44',
        "purple" : '45',
        "cyan" : '46',
        "white" : '47',
        "none": '0'
    }
    style = {
        'none' : 0,
        'bold' : 1,
        'underline' : 2,
        'italics' : 3,
        'underline' : 4,
    }

    def get_code(style='', fg='', bg=''):
        """Get ANSI color code:

        style;foreground;background

        Args:
            style ([type]): [description]
            fg ([type]): [description]
            bg ([type]): [description]

        Returns:
            [type]: [description]
        """
        ansi_code = '\033['
        code = ansi_code

        if style:
            code += str(style)
        if fg:
            code += f";{str(fg)}"
        if bg:
            code += f";{str(bg)}"
        if code == ansi_code:
            code += '0;0;0'

        code +='m'

        return code


class Config(object):
    
    @staticmethod
    def init(*paths: str, how: str = '', verbose: bool = False):
        if verbose:
            if how == 'no-dump':
                logger.info('no configuration dump found')
            elif how == 'empty':
                logger.info('initialized empty config')
            elif how == 'dump':
                logger.info('parsed config dump')
            elif paths:
                logger.info('parsed config files:')
                for index, path in enumerate(paths):
                    logger.info(f"[{str(index + 1)}] {path}")

    @staticmethod
    def get_section(section, verbose: bool = False):
        """Print the selected section.

        Args:
            section (Section): the selected section.
            verbose (bool, optional): Print iff `True`. Defaults to `False`.
        """        
        if verbose:
            print(json.dumps(section.get_values(), indent=4, sort_keys=True))


    @staticmethod
    def str(config):
        result: Dict[str, Dict[str, str]] = {}
        sections: List[str] = config.list_sections()

        for section in sections:
            result = { **result, **config.get_section(section).get_values() }

        return json.dumps(result, indent=4, sort_keys=True)


class Section(object):
    
    @staticmethod
    def init(*paths: str, verbose: bool = False):
        if verbose:
            pass # doesn't apply

    @staticmethod
    def get_values(values: Dict[str, Dict[str, str]], verbose: bool = False):
        """Print the selected key values.

        Args:
            values (Dict[str, Dict[str, str]]): the selected key values.
            verbose (bool, optional): Print iff `True`. Defaults to `False`.
        """        
        if verbose:
            print(json.dumps(values, indent=4, sort_keys=True))



"""_summary_
"""
from typing import Dict, Union, List
import lifespline_utils.utils as utils
import lifespline_utils.log as log
import json
from lifespline_utils.error import LSUInvalidArgumentsError, LSUInvalidSectionKeyError, LSUInvalidSectionError
import os
import functools as F


class Section(object):
    """`lifespline_utils.config.Section` is a config section.

    Internally, it is simple a dictionary wrapper. Each section is parsed into 
    a dictionary. The section:

    ```cfg
    [section]
    key = value
    ```

    Is parsed into:

    ```json
    {
        "section": {
            "key": "value"
        }
    }
    ```

    Which is wrapped by `lifespline_utils.config.Section` such that:
    ```json
    >>> config.get_sections('section').get_values()
    {
        "section": {
            "key": "value"
        }
    }
    ```
    """
    _name: str
    _values: Dict[str, str]
    
    def __init__(self, name: str, values: Dict[str, Union[str, List[str]]], verbose=False):
        """Initialize a new `lifespline_utils.config.Section`.

        Args:
            name (str): The section name
            values (Dict[str, Union[str, List[str]]]): The section values.
            verbose (bool, optional): Print to `stdout` iff `True`. Defaults to `False`.
        """        
        self._name = name
        self._values = values

        log.Section.init(verbose=verbose)

    def get_values(self, *keys: str, verbose: bool = False):
        """Get the values of a config section. See the examples below.

            Prints the selected key values.

        Args:
            keys (`(str)`): The selected section keys
            verbose (`bool`): Print the returned value to the `stdout`
            value_only (`bool`): Return only the value if `value_only`

        Returns:
            (`Dict[str, Dict[str, str]]`) The selected section keys

        Raises:

            (`lifespline_utils.error.LSUInvalidSectionKeyError`): The section
            key does not exist

        Examples

        ```python
        >>> config: Config = Config("test/test_config/setup1.cfg", verbose=True)
        >>> config.get_sections('section-1')
        {
            "section-1": {
                "key-1": "value-1",
                "key-2": "value-2",
                "key-3": "value-3"
            }
        }
        >>> config.get_sections('section-1').get_values()
        {
            "section-1": {
                "key-1": "value-1",
                "key-2": "value-2",
                "key-3": "value-3"
            }
        }

        >>> config.get_sections('section-1').get_values('key-1', 'key-3')
        {
            "section-1": {
                "key-1": "value-1",
                "key-3": "value-3"
            }
        }

        >>> config.get_sections('section-1').get_values('key-1')
        {
            "section-1": {
                "key-1": "value-1",
            }
        }

        >>> config.get_sections('section-1').get_values('key-4')
        ...
        lifespline_utils.error.LSUInvalidSectionKeyError: No such section key: key-4
        ```
        """
        values: Dict[str, Dict[str, str]] = {
            self._name: {}
        }

        if len(keys) == 0:
            values[self._name] = self._values
        elif len(keys) == 1:
            key: str = keys[0]
            value: str = self._values.get(key)
            if not value:
                raise LSUInvalidSectionKeyError(f"No such section key: {key}")

            values[self._name] = { key: value }
        elif len(keys) > 1:
            section_values: List[str] = []
            for key in list(keys):
                value: str = self._values.get(key)
                if not value:
                    err: str = f"No such section key: {key}"
                    raise LSUInvalidSectionKeyError(err)

                section_values.append((key, value))

            def to_values_dict(acc, pair):
                """Get a dictionary with keys `key1`, `key2` and the 
                corresponding values. 

                Args:
                    key1 (`str`): _description_
                    key2 (`str`): _description_

                Returns:
                    `Dict[str, str]`: A dictionary with keys `key1`, `key2` and the corresponding values.
                """
                key: str = pair[0]
                value: str = pair[1]

                res: Dict[str, str] = { **acc, **{ key: value } }

                return res

            acc: Dict = {}
            values[self._name] = F.reduce(to_values_dict, section_values, acc)

        log.Section.get_values(values, verbose=verbose)

        return values


    def __str__(self):
        """_summary_

        Returns:
            _type_: _description_
        """
        # TODO: move to lifespline_utils.log
        return json.dumps({self._name: self._values}, indent=4, sort_keys=True)

class Config(object):
    """`lifespline_utils.config.Config` is an object holding project configuration values.

        The object can be persisted in between contexts.
    """
    # config sections
    _sections: Dict[str, Section] = {}

    def __init__(self, *paths: str, dump: str = 'setup.json', verbose: bool = False):
        """Initialize a new `lifespline_utils.config.Config`.

            Loads configuration dump (`dump` or `setup.json`) if there is a 
            configuration dump. Otherwise it parses the specified config files.
            If there no paths are provided, it initializes an empty config.
            Prints the loaded configuration (if `verbose`).

            No section name sorting is guaranteed.

        # Args

            paths (`str`): The configuration files paths
            dump (`str`, optional): The configuration dumped file path. 
            Defaults to `setup.json`, i.e., the configuration is dumped to 
            your project root directory.
            verbose (`bool`, optional): `True` for verbose logging. Defaults to 
            `False`.

        # Returns

            `lifespline_utils.config.Config` A configuration object

        # Examples

        Load config from configuration dump

        ```python
        >>> config = Config(verbose=True)
        parsed config dump
        ```

        Create config without configuration dump and configuration files

        ```python
        >>> config = Config(verbose=True)
        initialized an empty config
        ```

        Create config from configuration paths

        ```python
        >>> paths = [
            "test/test_config/setup1.cfg",
            "test/test_config/setup2.cfg"
        ]
        >>> config = Config(*paths, verbose=True)
        parsed config files:
        [1] test/test_config/setup1.cfg
        [2] test/test_config/setup2.cfg

        >>> config
        {
            "section-1": {
                "key-1": "value-1",
                "key-2": "value-2",
                "key-3": "value-3"
            },
            "section-2": {
                "key-2": "value-2"
            }
        }
        ```
        """
        # Load configuration dump. 
        loaded: bool = self.load(path=dump, verbose=verbose)
        if not loaded:
            log.Config.init(how='no-dump', verbose=verbose)

            # If there is no configuration dump, read and parser the configuration files.
            if paths:
                config: Dict[str, Dict[str, str]]
                config = utils.parse_files(*list(paths))

                self._sections = Config._dict_to_sections(config)

                log.Config.init(*list(paths), verbose=verbose)

            if not paths:
                log.Config.init(how='empty', verbose=verbose)


    def __str__(self):
        """The string representation.
        """
        log.Config.str(self)

    @staticmethod
    def _dict_to_sections(config: Dict[str, Dict[str, str]]):
        """Convert dictionary sections to a dictionary of Sections

        # Args

            config (Dict[str, Dict[str, str]]): A config dictionary

        # Returns

            `Dict[str, lifespline_utils.config.Section]` Sections
        """
        sections: Dict[str, Section] = {}
        for section, values in config.items():
            sections[section] = Section(
                name=section,
                values=values
            )

        return sections


    def get_section(self, section: str, verbose: bool = False):
        """Get the config `section`.

            Prints the config section values (if `verbose`).

        Args:

            section `str`: The config section.

        Returns:

            `lifespline_utils.config.Section`: The config section.

        Raises:

            `lifespline_utils.error.LSUInvalidSectionError`: The provided section name does not exist in the config.

        Examples

        List sections contents
        ```python
        >>> config.get_sections('foo', verbose=True)
        ...
        lifespline_utils.error.LSUInvalidArgumentsError: No such section: 'foo'

        >>> config.get_sections('section-1', verbose=True)
        {
            "section-1": {
                "key-1": "value-1",
                "key-2": "value-2",
                "key-3": "value-3"
            }
        }
        ```
        """
        res: Section = self._sections.get(section)

        # the user specified an invalid section
        if not res:
            err: str = f"No such section: '{section}'"
            raise LSUInvalidSectionError(err)

        log.Config.get_section(section=res, verbose=verbose)

        return res


    def list_sections(self):
        """List the config section names.

            Prints the config section names.

        Returns:
            `List[str]` The list of section names
        """
        sections: List[str] = list(self._sections.keys())
        # return json.dumps(sections, indent=4, sort_keys=True)
        return sections


    def load(self, path: str = 'setup.json', verbose: bool = False):
        """Load a configuration dump.

        # Args

            path (`str`, optional):  The configuration dumped file path. 
            Defaults to `setup.json`, i.e., the configuration is dumped to 
            your project root directory.

        # Returns

            (`bool`): `True` if there was a valid config to be loaded, `False` otherwise.
        """
        res: bool = False

        config: Dict[str, Dict[str, str]]
        
        if os.path.exists(path):

            with open(path, 'r') as inp:
                config = json.load(inp)

            sections: Dict[str, Section]
            sections = Config._dict_to_sections(config)

            self._sections = sections.copy()

            res = True

        return res



from typing import Dict, List
import json
from lifespline_utils.config import Config, Section
from configparser import ConfigParser
from lifespline_utils.error import (
    LSUInvalidSectionKeyError,
    LSUInvalidSectionError
)

def test_get_section():
    """Test whether a section in a test file is correctly read as a config.
    Section.
    """
    assert True


def test_get_values():
    """Test whether a section values in a test file is correctly read.
    """
    assert True

# Test `lifespline_utils.config.Config.__init__`
# 
# Features to test
# - loads a config dump
# - read configuration files

def test_dict_to_config():
    """Test that a dictionary data structure is properly converted to a 
    dictionary of `lifespline_utils.config.Section`.
    """
    res: bool = True

    # create mock dump data file
    mock: Dict[str, Dict[str, str]]
    mock = {
        "section-1": {
            "key-1": "value-1",
            "key-2": "value-2",
            "key-3": "value-3"
        },
        "section-2": {
            "key-1": "value-1"
        },
    }

    # convert the dictionary to a dictionary of sections
    sections: Dict[str, Section]
    sections = Config._dict_to_sections(config=mock)

    # compare each section with the mock data
    for section, values in mock.items():
        mock_section: Dict[str, str] = mock[section]
        config_section: Section = sections[section]

        if res:

            for key in values.keys():
                mock_value: str = mock_section[key]
                section_value: str = config_section.get_values()[section][key]
                res = res and mock_value == section_value

                if not res:
                    break

    assert res


def test_load():
    """Verify that a dumped configuration file is properly loaded.

    Tests depending on this test:

    + `lifespline_utils.test.test_config.test_init_load_dump`.
    """
    res: bool = True

    # read test dump data file
    expected: Dict[str, Dict[str, str]]
    dump: str = 'test/test_config/setup.json'
    with open(dump) as json_file:
        expected = json.load(json_file)

    # load the configuration from the dumped file
    config: Config = Config(dump=dump)

    # compare 
    for section, values in expected.items():
        for key, value in values.items():
            expected_value: str = expected[section][key]
            value: str = config.get_section(section).get_values(key)[section][key]
            comparison = expected_value == value

            res = res and comparison

            if not res:
                assert expected_value == value

    assert res


def test_get_section_pass_argument():
    """Test whether `lifespline_utils.config.Config.get_section` properly 
    returns the requested section.

    Tests depending on this test:

    + `lifespline_utils.test.test_config.test_init_single_path`.
    """
    res: bool = True

    # read expected
    expected: Dict[str, Dict[str, str]]
    root: str = 'test/test_config/'
    setup: str = f'{root}setup1.cfg'
    parser: ConfigParser = ConfigParser()
    parser.read(setup)
    expected = parser._sections.copy()

    # create config from expected
    config: Config = Config(setup, verbose=True)

    # verify all section names exist in the config. If the section doesn't 
    # exist, an exception will be raised.
    for section in expected.keys():
        config.get_section(section)

    assert res


def test_get_section_pass_no_valid_section():
    """Test whether `lifespline_utils.config.Config.get_section` raises `lifespline_utils.error.LSUInvalidArgumentsError` if an invalid section name is passed as an argument.

    Assumes `lifespline_utils.config.Section.get_values`
    """
    res: bool = False

    # an empty config will not have a section name
    invalid_section_name: str = 'foo'
    config: Config = Config(verbose=True)
    try:
        config.get_section(invalid_section_name, verbose=True)
    except LSUInvalidSectionError:
        res = True

    assert res


def test_init_single_path():
    """Verify that `lifespline_utils.config.Config` is correctly initialized from a single configuration path.

    The test is covered by `lifespline_utils.test.test_config.
    test_get_section_pass_argument`. Reading from a single configuration file 
    path and checking for the expected sections is sufficient to prove the test.
    """
    pass 


def test_init_multiple_paths():
    """Verify that `lifespline_utils.config.Config` is correctly initialized from multiple configuration paths.
    """
    res: bool = True

    # read expected configuration files and parse them into a dictionary
    expected: Dict[str, Dict[str, str]] = {}
    root: str = 'test/test_config/'
    setups: List[str] = [
        f'{root}setup1.cfg',
        f'{root}setup2.cfg'
    ]
    parser: ConfigParser
    for setup in setups:
        parser = ConfigParser()
        parser.read(setup)
        expected = { **expected, **parser._sections.copy() }

    # init config from expected configuration files
    config: Config = Config(*setups, verbose=True)

    # compare the dictionary with the config by verifying all section names 
    # exist in the config. If the section doesn't exist, an exception will be 
    # raised. See: `lifespline_utils.config.Config.get_section`
    for section in expected.keys():
        config.get_section(section)

    assert res


def test_init_load_dump():
    """Verify that `lifespline_utils.config.Config` is correctly initialized from configuration dump.


    Test equivalent to `lifespline_utils.test.test_config.test_load`. The configuration does nothing more than loading the dump file.
    """
    pass 


def test_init_conflicting_config_sections():
    """Verify that `lifespline_utils.config.Config` raises an error if the 
    configuration files have a section header in common.
    """
    pass


def test_get_values_single_key():
    """Test whether `lifespline_utils.config.Section.get_values` properly 
    returns a correct single key value.

    Because this test is run on all keys, it also tests whether a config is properly initialized from a single config path.
    """
    res: bool = True

    # read expected
    expected: Dict[str, Dict[str, str]]
    root: str = 'test/test_config/'
    setup: str = f'{root}setup1.cfg'
    parser: ConfigParser = ConfigParser()
    parser.read(setup)
    expected = parser._sections.copy()

    # create config from expected
    config: Config = Config(setup, verbose=True)

    # compare section key value by key value
    for section, values in expected.items():
        for key, expected_value in values.items():
            true_value: str = config.get_section(section).get_values(key)
            true_value = true_value[section][key]
            comparison = expected_value == true_value

            res = res and comparison

            if not res:
                assert expected_value == true_value

    assert res


def test_get_values_multiple_keys():
    """Test whether `lifespline_utils.config.Config.get_section` properly 
    returns the requested section.
    """
    res: bool = True

    # read expected
    expected: Dict[str, Dict[str, str]]
    root: str = 'test/test_config/'
    setup: str = f'{root}setup1.cfg'
    parser: ConfigParser = ConfigParser()
    parser.read(setup)
    expected = parser._sections.copy()

    # create config from expected
    config: Config = Config(setup, verbose=True)

    # compare section key value by key value
    for section, values in expected.items():
        expected_keys: List[str] = list(values.keys())
        expected_values: List[str] = list(values.values())
        true_section: Dict[str, Section] = config.get_section(section)
        true_values: Dict[str, str] = true_section.get_values(*expected_keys)
        true_values: List[str] = list(true_values[section].values())

        expected_values.sort()
        true_values.sort()
        comparison: bool = expected_values == true_values

        if not comparison:
            assert expected_values == true_values

    assert res


def test_get_values_invalid_key():
    """Test whether `lifespline_utils.config.Config.get_section` properly 
    returns the requested section.
    """
    res: bool = False

    # the mock config setup does not have key-4
    root: str = 'test/test_config/'
    setup: str = f'{root}setup1.cfg'
    section: str = 'section-1'
    invalid_key: str = 'key-4'
    config: Config = Config(setup, verbose=True)

    # the line below must raise an exception for an unexisting key
    try:
        config.get_section(section).get_values(invalid_key, verbose=True)
    except LSUInvalidSectionKeyError as err:
        res = True

    assert res





"""_summary_."""

class LSUError(Exception):
    pass

class LSUInvalidArgumentsError(LSUError):
    """_summary_

    Args:
        LSUError (_type_): _description_
    """
    message: str

    def __init__(self, message):
        super().__init__(message)

class LSUInvalidSectionError(LSUError):
    """_summary_

    Args:
        LSUError (_type_): _description_
    """
    message: str

    def __init__(self, message):
        super().__init__(message)

class LSUInvalidSectionKeyError(LSUError):
    """_summary_

    Args:
        LSUError (_type_): _description_
    """
    message: str

    def __init__(self, message):
        super().__init__(message)


class LSUDuplicateConfigSectionsError(LSUError):
    """_summary_

    Args:
        LSUError (_type_): _description_
    """
    message: str

    def __init__(self, message):
        super().__init__(message)


class LSUNoSuchPathError(LSUError):
    """_summary_

    Args:
        LSUError (_type_): _description_
    """
    message: str

    def __init__(self, message):
        super().__init__(message)

class LSUNoSuchConfigFileError(LSUError):
    """No such configuration file

    Args:
        LSUError (_type_): _description_
    """
    message: str

    def __init__(self, message):
        super().__init__(message)

class LSUInvalidConfigFileFormatError(LSUError):
    """Invalid configuration file format

    Args:
        LSUError (_type_): _description_
    """    
    message: str

    def __init__(self, message):
        super().__init__(message)

class LSUPythonVirtualEnvironmentNotFoundError(LSUError):
    """The python virtual environment `.env` was not found.

    Args:
        LSUError (_type_): _description_
    """    
    message: str

    def __init__(self, message):
        super().__init__(message)



import os
from pathlib import Path
from lifespline_utils.log import logger
import pathlib
from lifespline_utils.error import (
    LSUInvalidConfigFileFormatError,
    LSUNoSuchPathError,
    LSUDuplicateConfigSectionsError
)
from typing import List, Dict
import configparser

# TODO mv to config?
_project_name: str = 'lifespline-utils'

def parse_files(*paths: str):
    """Parse files of type `.cfg` into dictionary data structure.

    Validates whether or not each path in `paths` is a well-structure cfg file, 
    and parses the files into a map of dictionaries if they're valid.

    :raises:
        `lifespline_utils.error.LSUInvalidFileFormatError`: The cfg file does not have the correct format.

        `lifespline_utils.error.LSUDuplicateConfigSectionsError`: Two 
        configuration files have the same section name.

    Args:
        path (_type_): _description_
    """

    def have_duplicate_keys(cfg1, cfg2):
        """Check if dicts `cfg1` and `cfg2` have duplicate keys.
        headers.

        Args:
            cfg1 (`dict`): dictionary
            cfg2 (`dict`): dictionary

        Returns:
            (`bool`): `True` if there are duplicate keys, `False` otherwise.
        """
        cfg1_keys = set(cfg1)
        cfg2_keys = set(cfg2)
        cfgs_keys = set({ **cfg1, **cfg2 })

        return len(cfg1_keys) + len(cfg2_keys) > len(cfgs_keys)

    # configuration file parser
    parser: configparser.ConfigParser

    # dict with parsed configuration files
    config: Dict[str, Dict[str, str]] = {}

    # get absolute paths
    abs_paths: List[str] = [get_path(path) for path in paths]

    # read and parse config files, add all parsed configuration files to
    # a dict data structure
    for abs_path in abs_paths:
        try:
            parser = configparser.ConfigParser()
            parser.read(abs_path)

            if have_duplicate_keys(cfg1=parser._sections, cfg2=config):
                err: str = f"Found duplicate keys in '{abs_path}' and some other configuration file."
                raise LSUDuplicateConfigSectionsError(err)

            config = { **config, **parser._sections }

        except configparser.MissingSectionHeaderError:
            err: str = f"Missing section header in '{abs_path}'"
            raise LSUInvalidConfigFileFormatError(err)

    return config

def get_project_root(verbose: bool = True, venv: str = '.env'):
    """Get the absolute project root path.

        Get the absolute path of your project. This is achieved by getting the parent directory of your project's python virtual environment, which defaults to `.env`.

        The project root is important to get absolute paths for whatever files you might be interested in finding whithin your project.

    ## Parameters

    + verbose (str, optional): [description]. Defaults to None.
    + venv (str, optional): The name of the python virtual environment. 
    Defaults to '.env'.

    ## Returns

        `str`: The project root path.

    ## Examples

    Having installed the package in development mode (from src)
    ```bash
    >>> from lifespline_utils.utils import get_project_root
    >>> get_project_root()
    "{home}/{lifespline-utils-project-path}/src/lifespline_utils/"
    ```

    Having installed the package from PyPi/build
    ```bash
    >>> from lifespline_utils.utils import get_project_root
    >>> get_project_root()
    "{home}/{your-project-path}/"
    ```
    """  
    curr_file_path: str = str(pathlib.Path(__file__).resolve())

    try:
        fst_index: int = curr_file_path.index(venv)
    except ValueError:
        # if the virtual environment could not be found, then the package
        # is running from a local clone.
        fst_index: int = curr_file_path.index(_project_name)
        lst_index: int = fst_index + len(_project_name) + 1
        fst_index = lst_index

    project_root: str = curr_file_path[:fst_index]

    return project_root

def get_path(path: str = ''):
    """Get the absolute path for `file`.

    Get the absolute path for `file` if `file` exists in the user home dir or 
    in the project root (see `lifespline_utils.utils.get_project_root`). `file` 
    can be either a file or a directory.

    ## Parameters

    + file (str, optional): [description]. Defaults to None.
    + verbose (str, optional): [description]. Defaults to None.

    ## Returns

        (`str`) The file path if the path exists, None if it doesn't exist.

    ## Examples

    Having installed the package in development mode (from src)
    ```bash
    >>> from lifepsline_utils.utils import get_path
    >>> get_path()
    "/{home}/{lifespline-utils-project-path}/src/lifespline_utils/utils.py"
    >>> get_path("README.md")
    "/{home}/{lifespline-utils-project-path}/README.md"
    >>> get_path(".bashrc")
    "/{home}/.bashrc"
    ```

    When installing from the PyPi/build
    ```bash
    >>> from lifepsline_utils.utils import get_path
    >>> get_path()
    '/{home}/{your-project-venv-package-egg-path}/lifespline_utils/utils.py'
    >>> get_path("README.md")
    "/{home}/{your-project-path}/README.md"
    >>> get_path(".bashrc")
    "/{home}/.bashrc"
    ```

    ```bash
    >>> get_path('foo')
    "LSUNoSuchPathError: '/home/{your-project-path}/foo'"
    ```
    """
    root: str
    file_path: str = str(pathlib.Path(__file__).resolve())

    if path:
        root = get_project_root()
        file_path = os.path.join(root, path)
        # verify whether the path exists, otherwise, verify if exists in the 
        # user home directory
        if not os.path.exists(file_path):
            root = os.path.expanduser('~')
            file_path = os.path.join(root, path)
            if not os.path.exists(file_path):
                raise LSUNoSuchPathError(f"No such path: '{file_path}'")

    return file_path
